# -*- coding: utf-8 -*-
import uno  # pragma: no flakes
from com.sun.star.script.provider import XScriptContext

from unotools.unohelper import ContextBase


class LocalContext(ContextBase):

    def __init__(self, context=None):
        super().__init__(context)
        self.resolver = self.create_resolver(self.raw)


class ScriptContext(ContextBase, XScriptContext):
    pass



# -*- coding: utf-8 -*-
import argparse
import logging
from os.path import isdir, isfile

try:
    from functools import singledispatch
except ImportError:  # under 3.3
    from singledispatch import singledispatch

from unotools.context import LocalContext, ScriptContext
from unotools.errors import ArgumentError, ConnectionError


class Pipe:
    def __init__(self, name: str):
        self.name = name


class Socket:
    def __init__(self, host: str, port: int):
        self.host = host
        self.port = port


@singledispatch
def connect(identifier: str, **kwargs) -> ScriptContext:
    context = None
    try:
        local_context = LocalContext()
        option = kwargs.get('option')
        if option is not None:
            identifier = '{},{}'.format(identifier, option)
        conn_str = 'uno:{};urp;StarOffice.ComponentContext'.format(identifier)
        context = local_context.resolver.resolve(conn_str)
        if context:
            return ScriptContext(context)
    except Exception as e:
        msg = 'failed to connect: {}'.format((identifier, kwargs))
        raise ConnectionError(msg) from e

    raise ConnectionError('cannot connect: {}'.format((identifier, kwargs)))


@connect.register(Pipe)
def connect_with_pipe(pipe: Pipe, **kwargs) -> ScriptContext:
    return connect('pipe,name={}'.format(pipe.name), **kwargs)


@connect.register(Socket)
def connect_with_socket(socket: Socket, **kwargs) -> ScriptContext:
    identifier = 'socket,host={},port={}'.format(socket.host, socket.port)
    return connect(identifier, **kwargs)


def parse_argument(argv: list):
    """
    >>> argv = '-s server -p 8080 -a tcpNoDelay=1'.split()
    >>> parse_argument(argv)  # doctest: +NORMALIZE_WHITESPACE
    Namespace(datadirs=[], encoding='utf-8', file_=None, host='server',
              option='tcpNoDelay=1', outputdir='.', pipe=None, port=8080,
              verbose=False)
    """
    def validate_file_or_directory_existence(args):
        def validate(paths, validate_func):
            for path in paths:
                if path is not None and not validate_func(path):
                    raise ArgumentError('"{}" is not found'.format(path))
        validate([args.file_], isfile)
        validate(args.datadirs + [args.outputdir], isdir)

    parser = argparse.ArgumentParser()
    parser.set_defaults(datadirs=[], encoding='utf-8', file_=None,
                        host='localhost', option=None, outputdir='.',
                        pipe=None, port=8100, verbose=False)
    parser.add_argument('-a', '--option', dest='option',
                        metavar='OPTION', help='set option')
    parser.add_argument('-d', '--datadirs', dest='datadirs', nargs='*',
                        metavar='DATADIRS', help='set data directories')
    parser.add_argument('-e', '--encoding', dest='encoding',
                        metavar='ENCODING',
                        help='set encoding. default is utf-8')
    parser.add_argument('-f', '--file', dest='file_',
                        metavar='FILE', help='set OpenDocument file to read')
    parser.add_argument('-i', '--pipe', dest='pipe',
                        metavar='PIPE', help='set pipe name')
    parser.add_argument('-o', '--outputdir', dest='outputdir',
                        metavar='OUTPUTDIR', help='set output directory')
    parser.add_argument('-p', '--port', dest='port', type=int,
                        metavar='PORT_NUMBER', help='set port number')
    parser.add_argument('-s', '--host', dest='host', required=True,
                        metavar='HOST', help='set host name')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='set verbose mode')
    args = parser.parse_args(argv)
    validate_file_or_directory_existence(args)

    if args.verbose:
        logging.root.setLevel(logging.DEBUG)
    logging.debug('args: {0}'.format(args))
    return args



# -*- coding: utf-8 -*-
import sys
from os.path import join as pathjoin

from unotools import Socket, connect
from unotools.unohelper import constant
from unotools.unohelper import convert_path_to_url
from unotools.component.calc import Calc
from unotools.component.writer import Writer
from unotools.data.csv import CsvFile
from unotools.utils import search_file


def calc_sample(args, context):
    calc = Calc(context)
    sheet = calc.get_sheet_by_index(0)
    sheet.set_columns_str(0, 0, ['Month', 'Sales', 'End Date'])

    path = next(search_file(args.datadirs[0], 'data1.csv'))
    csv_file = CsvFile(path, has_header=True)
    for i, data in enumerate(csv_file.read(), 1):
        sheet.set_rows(0, i, data, csv_file.header)

    format_date = constant('com.sun.star.util.NumberFormat.DATE')
    formats = calc.get_number_formats()
    locale = context.create_struct('com.sun.star.lang.Locale')
    cell_range = sheet.get_cell_range_by_name('C2:C13')
    cell_range.NumberFormat = formats.getStandardFormat(format_date, locale)

    chart_cell_range = sheet.get_cell_range_by_name('A1:B13')
    sheet.add_charts_new_by_name('Sales',
                context.make_rectangle(8000, 1000, 16000, 10000),
                chart_cell_range.get_range_address(), True, True)
    chart = sheet.get_chart_by_name('Sales')
    chart_doc = chart.get_embedded_object()

    title_text_shape = chart_doc.get_title()
    title_text_shape.String = 'Sales Chart'

    diagram = chart_doc.create_instance('com.sun.star.chart.BarDiagram')
    diagram.Vertical = True
    diagram.DataCaption = constant('com.sun.star.chart.ChartDataCaption.VALUE')
    chart_doc.set_diagram(diagram)

    sheets_count = calc.get_sheets_count()
    new_sheets_data = ['sales', 'benefit', 'budget']
    calc.insert_multisheets_new_by_name(new_sheets_data, sheets_count)
    calc.get_sheet_by_name('budget').set_name('cost')

    base_path = convert_path_to_url(pathjoin(args.outputdir, 'sample'))
    calc.store_to_url(base_path + '.xls', 'FilterName', 'MS Excel 97')
    calc.store_to_url(base_path + '.csv',
                      'FilterName', 'Text - txt - csv (StarCalc)')
    calc.store_to_url(base_path + '-calc.pdf', 'FilterName', 'calc_pdf_Export')
    calc.store_to_url(base_path + '-calc.html',
                      'FilterName', 'HTML (StarCalc)')

    calc.close(True)


if __name__ == '__main__':
    from unotools import parse_argument
    args = parse_argument(sys.argv[1:])
    context = connect(Socket(args.host, args.port), option=args.option)
    calc_sample(args, context)



# -*- coding: utf-8 -*-
import functools
import glob
import os
from os.path import join as pathjoin


def cached_property(func: callable) -> property:
    return property(functools.lru_cache()(func))


def get_annotation_to_kwargs(klass: type, func_name: str, values: dict
                             ) -> dict:
    ann = getattr(klass, func_name).__annotations__
    return {key: values.get(key) for key in ann.keys()
            if values.get(key) is not None}


def set_kwargs(obj: object, kwargs: dict):
    for key, value in kwargs.items():
        if value is not None:
            setattr(obj, key.title(), value)


def convert_lowercase_to_camecase(name: str) -> str:
    """
    >>> convert_lowercase_to_camecase('')
    ''
    >>> convert_lowercase_to_camecase('close')
    'close'
    >>> convert_lowercase_to_camecase('get_title')
    'getTitle'
    >>> convert_lowercase_to_camecase('set_array_value')
    'setArrayValue'
    >>> convert_lowercase_to_camecase('getURL')
    'getURL'
    """
    tokens = name.split('_')
    first, rest = tokens[0], tokens[1:]
    return first + ''.join(i.title() for i in rest)


def get_file_list(path: str) -> str:
    for root, dirs, files in os.walk(path):
        for filename in files:
            yield pathjoin(root, filename)


def search_file(path: str, name: str) -> str:
    for root, dirs, files in os.walk(path):
        for file_path in glob.glob(os.path.join(root, name)):
            yield file_path



# -*- coding: utf-8 -*-
from operator import methodcaller

import uno  # pragma: no flakes
from com.sun.star.awt import Rectangle
from com.sun.star.table import CellRangeAddress

from unotools.unohelper import ComponentBase
from unotools.unohelper import LoadingComponentBase
from unotools.datatypes import Sequence


class ChartDocument(ComponentBase):
    pass


class TableChart(ComponentBase):

    def get_embedded_object(self) -> ChartDocument:
        return ChartDocument(self.context, self.raw.getEmbeddedObject())


class TableCharts(ComponentBase):
    pass


class CellRange(ComponentBase):
    pass


class Cell(ComponentBase):
    pass


class Spreadsheet(ComponentBase):

    def set_rows_cell_data(self, x: int, y: int, data: list, method_name: str):
        for datum in data:
            methodcaller(method_name, datum)(self.raw.getCellByPosition(x, y))
            y += 1

    def set_columns_cell_data(self, x: int, y: int, data: list,
                              method_name: str):
        for datum in data:
            methodcaller(method_name, datum)(self.raw.getCellByPosition(x, y))
            x += 1

    def set_rows(self, x: int, y: int, data: list, method_names: str):
        for i, datum in enumerate(data):
            self.set_rows_cell_data(x, y, [datum], method_names[i])
            x += 1

    def set_columns(self, x: int, y: int, data: list, method_names: str):
        for i, datum in enumerate(data):
            self.set_rows_cell_data(x, y, [datum], method_names[i])
            y += 1

    def set_rows_str(self, x: int, y: int, data: list):
        self.set_rows_cell_data(x, y, data, 'setString')

    def set_columns_str(self, x: int, y: int, data: list):
        self.set_columns_cell_data(x, y, data, 'setString')

    def set_rows_value(self, x: int, y: int, data: list):
        self.set_rows_cell_data(x, y, data, 'setValue')

    def set_columns_value(self, x: int, y: int, data: list):
        self.set_columns_cell_data(x, y, data, 'setValue')

    def set_rows_formula(self, x: int, y: int, data: list):
        self.set_rows_cell_data(x, y, data, 'setFormula')

    def set_columns_formula(self, x: int, y: int, data: list):
        self.set_columns_cell_data(x, y, data, 'setFormula')

    def get_cell_by_position(self, x: int, y: int) -> Cell:
        return Cell(self.context, self.raw.getCellByPosition(x, y))

    def get_cell_range_by_name(self, range_: str) -> CellRange:
        return CellRange(self.context, self.raw.getCellRangeByName(range_))

    def get_cell_range_by_position(self, left: int, top: int, right: int,
                                   bottom: int) -> CellRange:
        raw = self.raw.getCellRangeByPosition(left, top, right, bottom)
        return CellRange(self.context, raw)

    @property
    def charts(self) -> TableCharts:
        return TableCharts(self.context, self.raw.getCharts())

    def get_charts_count(self) -> int:
        return self.charts.getCount()

    def add_charts_new_by_name(self, name: str,
                               rect: Rectangle, ranges: CellRangeAddress,
                               column_headers: bool, row_headers: bool):
        self.charts.addNewByName(name, rect, Sequence(ranges),
                                 column_headers, row_headers)

    def get_chart_by_index(self, index: int) -> TableChart:
        return TableChart(self.context, self.charts.getByIndex(index))

    def get_chart_by_name(self, name: str) -> TableChart:
        return TableChart(self.context, self.charts.getByName(name))


class Spreadsheets(ComponentBase):
    pass


class Calc(LoadingComponentBase):

    URL = 'private:factory/scalc'

    @property
    def sheets(self) -> Spreadsheets:
        return Spreadsheets(self.context, self.raw.getSheets())

    def get_sheets_count(self) -> int:
        return self.sheets.getCount()

    def get_sheet_by_index(self, index: int) -> Spreadsheet:
        return Spreadsheet(self.context, self.sheets.getByIndex(index))

    def get_sheet_by_name(self, name: str) -> Spreadsheet:
        return Spreadsheet(self.context, self.sheets.getByName(name))

    def insert_sheets_new_by_name(self, name: str, position: int):
        self.sheets.insertNewByName(name, position)

    def insert_multisheets_new_by_name(self, data: list, position: int):
        for i, datum in enumerate(data):
            self.insert_sheets_new_by_name(datum, position + i)

    def remove_sheets_by_name(self, name: str):
        self.sheets.removeByName(name)



# -*- coding: utf-8 -*-
import sys
from os.path import join as pathjoin

from unotools import Socket, connect
from unotools.unohelper import convert_path_to_url
from unotools.component.writer import Writer


def writer_sample(args, context):
    writer = Writer(context)
    writer.set_string_to_end('world\n')
    writer.set_string_to_start('hello\n')

    base_path = convert_path_to_url(pathjoin(args.outputdir, 'sample'))
    writer.store_to_url(base_path + '.doc', 'FilterName', 'MS Word 97')
    writer.store_to_url(base_path + '-writer.pdf',
                        'FilterName', 'writer_pdf_Export')
    writer.store_to_url(base_path + '-writer.html',
                        'FilterName', 'HTML (StarWriter)')

    writer.close(True)


if __name__ == '__main__':
    from unotools import parse_argument
    args = parse_argument(sys.argv[1:])
    context = connect(Socket(args.host, args.port), option=args.option)
    writer_sample(args, context)



# -*- coding: utf-8 -*-
from com.sun.star.text import XText
from com.sun.star.text import XTextRange

from unotools.unohelper import ComponentBase
from unotools.unohelper import LoadingComponentBase


class Text(ComponentBase):
    pass


class Writer(LoadingComponentBase):

    URL = 'private:factory/swriter'

    @property
    def text(self) -> XText:
        return Text(self.context, self.raw.getText())

    def get_start(self) -> XTextRange:
        return self.text.getStart()

    def get_end(self) -> XTextRange:
        return self.text.getEnd()

    def set_string_to_start(self, text: str):
        self.get_start().setString(text)

    def set_string_to_end(self, text: str):
        self.get_end().setString(text)



# -*- coding: utf-8 -*-
import csv
from os.path import basename


class CsvFile:

    def __init__(self, path: str, mode: str='r', encoding: str='utf-8',
                 has_header: bool=False, convert_func: callable=lambda x: x,
                 **kwargs):
        self.path = path
        self.file_name = basename(path)
        self.mode = mode
        self.encoding = encoding
        self.header = None
        self.has_header = has_header
        self.convert_func = convert_func
        self.reader_kwargs = kwargs

    def read(self) -> list:
        with open(self.path, mode=self.mode, encoding=self.encoding) as f:
            reader = csv.reader(f, **self.reader_kwargs)
            if self.has_header:
                self.header = next(reader)
            for data in reader:
                yield list(map(self.convert_func, data))



# -*- coding: utf-8 -*-
import sys
from setuptools import setup


VERSION = '0.3.2'

try:
    import pypandoc
    LONG_DESCRIPTION = '\n'.join([
        pypandoc.convert('README.md', 'rst'),
        pypandoc.convert('CHANGELOG.md', 'rst'),
    ])
except (IOError, ImportError):
    LONG_DESCRIPTION = ''

REQUIRES = []
if sys.version_info < (3, 4):
    REQUIRES.append('singledispatch')

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: POSIX',
    'Programming Language :: Python :: 3',
    'Topic :: Software Development',
    'Topic :: Office/Business :: Office Suites',
]

setup(
    name='unotools',
    version=VERSION,
    description='Interacting with OpenOffice.org/LibreOffice using UNO',
    long_description=LONG_DESCRIPTION,
    classifiers=CLASSIFIERS,
    keywords=['uno', 'pyuno', 'office', 'OOo'],
    author='Tetsuya Morimoto',
    author_email='tetsuya dot morimoto at gmail dot com',
    url='http://bitbucket.org/t2y/unotools',
    license='Apache License 2.0',
    platforms=['unix', 'linux', 'osx'],
    packages=['unotools'],
    include_package_data=True,
    install_requires=REQUIRES,
    tests_require=['tox', 'pytest', 'pytest-pep8', 'pytest-flakes'],
)



# -*- coding: utf-8 -*-
from os.path import realpath

import uno
import unohelper
import pyuno
from com.sun.star.bridge import XUnoUrlResolver
from com.sun.star.frame import XDesktop
from com.sun.star.lang import XComponent
from com.sun.star.reflection import XIdlReflection
from com.sun.star.uno import XComponentContext
from com.sun.star.uno import XInterface

from unotools.datatypes import Sequence
from unotools.utils import cached_property
from unotools.utils import convert_lowercase_to_camecase
from unotools.utils import get_annotation_to_kwargs
from unotools.utils import set_kwargs


class Base:

    def __getattr__(self, name: str) -> XInterface:
        if self.raw is not None:
            return getattr(self.raw, convert_lowercase_to_camecase(name))
        raise AttributeError

    def _show_attributes(self):
        from pprint import pprint
        pprint(dir(self.raw))


class ContextBase(Base, unohelper.Base):

    def __init__(self, context=None):
        self.raw = uno.getComponentContext() if context is None else context
        self.service_manager = self.raw.getServiceManager()
        self.core_reflection = self.create_core_reflection(self.raw)

    def as_raw(self) -> XComponentContext:
        return self.raw

    def create_instance(self, name: str) -> XInterface:
        return self.service_manager.createInstance(name)

    def create_instance_with_context(self, name: str,
                                     context: XComponentContext
                                     ) -> XInterface:
        return self.service_manager.createInstanceWithContext(name, context)

    def create_resolver(self, context: XComponentContext) -> XUnoUrlResolver:
        service_name = 'com.sun.star.bridge.UnoUrlResolver'
        return self.create_instance_with_context(service_name, context)

    def create_core_reflection(self, context: XComponentContext
                               ) -> XIdlReflection:
        service_name = 'com.sun.star.reflection.CoreReflection'
        return self.create_instance_with_context(service_name, context)

    def create_desktop(self, context: XComponentContext) -> XDesktop:
        service_name = 'com.sun.star.frame.Desktop'
        return self.create_instance_with_context(service_name, context)

    @cached_property
    def document(self, context: XComponentContext) -> XComponent:
        return self.create_desktop(context).getCurrentComponent()

    def create_simple_file_access(self, context: XComponentContext
                                  ) -> XIdlReflection:
        service_name = 'com.sun.star.ucb.SimpleFileAccess'
        return self.create_instance_with_context(service_name, context)

    def create_struct(self, type_name: str) -> uno.Any:
        rv, struct = self.core_reflection.forName(type_name).createObject(None)
        return struct

    def make_struct_data(self, type_name: str, **kwargs) -> uno.Any:
        struct = self.create_struct(type_name)
        set_kwargs(struct, kwargs)
        return struct

    def make_point(self, x: int, y: int) -> uno.Any:
        kwargs = self._get_kwargs('make_point', locals())
        return self.make_struct_data('com.sun.star.awt.Point', **kwargs)

    def make_property_value(self, name: str=None, value: str=None,
                            handle: int=None, state: object=None
                            ) -> uno.Any:
        type_name = 'com.sun.star.beans.PropertyValue'
        kwargs = self._get_kwargs('make_property_value', locals())
        return self.make_struct_data(type_name, **kwargs)

    def make_rectangle(self, x: int, y: int, width: int, height: int
                       ) -> uno.Any:
        kwargs = self._get_kwargs('make_rectangle', locals())
        return self.make_struct_data('com.sun.star.awt.Rectangle', **kwargs)

    def make_size(self, width: int, height: int) -> uno.Any:
        kwargs = self._get_kwargs('make_size', locals())
        return self.make_struct_data('com.sun.star.awt.Size', **kwargs)

    def _get_kwargs(self, func_name: str, values: dict) -> dict:
        return get_annotation_to_kwargs(self.__class__, func_name, values)


class ComponentBase(Base):

    def __init__(self, context: XComponentContext, component: XComponent):
        self.context = context
        self.raw = component

    def as_raw(self) -> XComponent:
        return self.raw

    def store_as_url(self, url: str, *values):
        self.raw.storeAsURL(url, self._get_property_values(*values))

    def store_to_url(self, url: str, *values):
        self.raw.storeToURL(url, self._get_property_values(*values))

    def _get_property_values(self, *values) -> Sequence:
        if len(values) == 1 and values[0] is None:
            return Sequence()
        else:
            return Sequence(self.context.make_property_value(*values))


class LoadingComponentBase(ComponentBase):

    def __init__(self, context: XComponentContext,
                 url: str=None,
                 target_frame_name: str='_blank',
                 search_flags: int=0,
                 arguments: tuple=()):
        self.context = context
        self.desktop = self.context.create_desktop(self.context.raw)
        self.url = url if url is not None else self.URL
        self.raw = self.desktop.loadComponentFromURL(self.url,
                                                     target_frame_name,
                                                     search_flags, arguments)


# alias
def constant(name: str) -> object:
    """
    >>> constant('com.sun.star.util.NumberFormat.DATE')
    2
    """
    return pyuno.getConstantByName(name)


def unotype(name: str) -> uno.Type:
    """
    >>> unotype('com.sun.star.uno.XInterface') # doctest: +NORMALIZE_WHITESPACE
    <Type instance com.sun.star.uno.XInterface
        (<uno.Enum com.sun.star.uno.TypeClass ('INTERFACE')>)>
    """
    return pyuno.getTypeByName(name)


def unoclass(name: str) -> type:
    """
    >>> unoclass('com.sun.star.text.XText')
    <class 'uno.com.sun.star.text.XText'>
    """
    return pyuno.getClass(name)


def convert_path_to_url(path: str) -> str:
    """
    >>> convert_path_to_url('/var/tmp/libreoffice')
    'file:///var/tmp/libreoffice'
    """
    return pyuno.systemPathToFileUrl(realpath(path))


def convert_url_to_path(url: str) -> str:
    """
    >>> convert_url_to_path('file:///var/tmp/libreoffice')
    '/var/tmp/libreoffice'
    """
    return pyuno.fileUrlToSystemPath(url)
